#ifndef WEAK_CLASSIFIER_H
#define WEAK_CLASSIFIER_H

#include "integral_image.h"
#include <vector>
#include <algorithm>
//#include <boost/archive/text_oarchive.hpp>
//#include <boost/archive/text_iarchive.hpp>

/**
 *
 *
 *
 *
 *
 * Time-stamp: <2005-08-03 19:05:20 engmb>
 * Copyright (C) 2005 by Manuel Boissenin
 *
 * @author Manuel Boissenin
 * @date 29/07/2005
 ***********************************************************************/
namespace mimas{

class corner_reference
{
 public:
  unsigned int x;
  unsigned int y;
  float weight;
  corner_reference(unsigned int x, unsigned int y, float w)
    :x(x), y(y), weight(w)
    {}

  /*
 private:
  friend class boost::serialization::access;
  
  template<class Archive>
    void serialize(Archive & ar, const unsigned int version)
    {
      ar & x;
      ar & y;
      ar & weight;
    }
  */
};


/**
Class to calculate an haar_feature according to the pattern defined by the constructor.  

The top left corner position of the haar_feature as well as a multiplicative factor can be specified. 

*/
class haar_feature
{
  //!!!put a static variable to tell the size of the considered window
  //(i.e. for face the dimension of the face)
 public:

  //!!!comment if not necesary
  float measure(integral_image &i_image)
  {
    float sum = 0.;
    for(it = corner_references.begin(); it != corner_references.end(); it++)
      {
	sum += it->weight * (i_image.getPixel(it->x, it->y));
      }
    return sum;
  }

  float measure(integral_image &i_image, int x_pos, int y_pos, float factor) throw(std::string) 
  {
    float sum = 0.;
    for(it = corner_references.begin(); it != corner_references.end(); it++)
      {
	if((factor * it->x + x_pos) < 0 || (factor * it->x + x_pos) > i_image.getWidth()||
	   (factor * it->y + y_pos) < 0 || (factor * it->y + y_pos) > i_image.getHeight())
	  {
	    throw std::string("integral image out of bound");
	  }
	  sum += it->weight * (i_image.getPixel( int((factor * it->x) + x_pos), int((factor * it->y) + y_pos)));
      }
    return sum;
  }

  
  void print(void)
    {
      for(it = corner_references.begin(); it != corner_references.end(); it++)
	std::cerr<<"x : " <<it->x <<"y : " <<it->y << "weight : " <<it->weight <<std::endl;
    }

  haar_feature(std::vector<corner_reference> &corner_ref)
    :corner_references(corner_ref)
    {}
  
 private:
  /*
  friend class boost::serialization::access;
  
  template<class Archive>
    void serialize(Archive & ar, const unsigned int version)
    {
      ar & corner_references;
    }
  */
    
  std::vector<corner_reference> corner_references;
  std::vector<corner_reference>::iterator it;

};

//1 means positive, -1 negative
//For storing the name of positve or negative examples
class ref_image {
 public:
  std::string image_name;
  double weight_associated;
  int status;  //to say if the example is positive or negative
  ref_image(std::string image_name, double weight_associated, int status)
    :image_name(image_name), weight_associated(weight_associated), status(status)
    {}
};

class feature_measure{
 public:
  ref_image *im_ref; //use a smart pointer
  float result;

  //check that ref_image has been allocated.
  feature_measure(ref_image *im_ref, float result)
    :im_ref(im_ref), result(result)
    {}
};

struct feature_measure_less_than
{
     bool operator()(const feature_measure& left, const feature_measure& right)
     {
          return left.result < right.result;
     }
};

typedef enum{POSITIVE_BELOW = -1 , POSITIVE_ABOVE = 1} direction_t;

class weak_classifier
{
 public:
  float threshold;
  direction_t threshold_direction;
  double error_measure;

  //!!redondant?
  static void add_image(ref_image *im_ref);    
  static void normalise_weight_and_calculate_sum(void);
  static void reweight(weak_classifier &best_wc)
    {
      double multiplication_factor = best_wc.error_measure / (1 - best_wc.error_measure); 

      int nb_reweight = 0;
      
      for(std::vector<feature_measure>::iterator it = best_wc.measures.begin(); it!= best_wc.measures.end(); it++)
	{
	  //in the case where we should be with a negative example
	  if((best_wc.threshold_direction * it->result) < (best_wc.threshold_direction *  best_wc.threshold))	   
	    //if we have it we reweight the image giving a lower weight
	    {
	      if(it->im_ref->status == -1)
		{
		  it->im_ref->weight_associated *= multiplication_factor;
		  nb_reweight ++;
		}
	      else
		{
		  //std::cerr<<"positive example image that mismatch the classifier: "<< it->im_ref->image_name<<std::endl;
		}
	    }
	  //in the case we should have a positive example
	  else
	    {
	      if(it->im_ref->status == 1)
		{
		  it->im_ref->weight_associated *= multiplication_factor;
		  nb_reweight ++;
		}
	      else
		{
		  //std::cerr<<"negative example image that mismatch the classifier: "<< it->im_ref->image_name<<std::endl;
		}
	    }	  
	}

      for(std::vector<feature_measure>::iterator it = best_wc.measures.begin(); it!= best_wc.measures.end(); it++)
	{
	  if(it->im_ref->weight_associated < 0)
	    std::cerr<<"ALARM"<<std::endl;
	}
      std::cerr << "Number of reweight: "<< nb_reweight<< std::endl;
    }

  void sort_measures(void)
    {
      sort(measures.begin(), measures.end(), feature_measure_less_than());      
    }

  //find the threshold value as well as the error measure
  //for a given set of weight of the images
  void determine_threshold(void);

  haar_feature feature;
  std::vector<feature_measure> measures;
  
  weak_classifier(haar_feature &feat)
    :feature(feat){}
  
  void measure(ref_image *im_ref, integral_image &i_image)
    {
      measures.push_back(feature_measure(im_ref, feature.measure(i_image)));
    }

 private:
  static std::vector<ref_image *> image_references;
  
  static double total_sum_weight_positive_example;
  static double total_sum_weight_negative_example;
};


}

#endif
