/**********************************************************************
*example code how to use the hypothesis filter.
*Usage of the partial reinitialisation.
*
*
*
*
* @author: Manuel Boissenin
* @date: 20/05/2004
*
* Time-stamp: <2005-05-20 12:53:48 engmb>
* Copyright (C) 2004 by Manuel Boissenin
*
 ***********************************************************************/

#include <string>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <fstream>
#include <image.h>
#include <image_fileinput.h>
#include <image_fileoutput.h>
#include <GL/glut.h>
#include <image_mesaoutput.h>

//!!change the name of p_condensation p stands for partial reinitialisation

#include "p_condensation.h"
#include "hypothesis2D.h"
#include "correlation_observer2D.h"
#include "analyse_result_blob.h"
#include "image_loader.h"
#include "compensate_speed.h" //why (and how: included from hypothesis_filter.h) am I using no_compensate then?!!!!
//do another example with compensate speed !!?

using namespace mimas::hf;
using namespace std;

int main(int argc, char** argv)
{
  int retVal = 0;
  try {
    glutInit(&argc, argv);
    mimas::x11_display disp; 
    image_mesaoutput< unsigned char > display(&disp);

    //Initialisation of the tracker
    hypothesis2D init(atoi(argv[2]),atoi(argv[3]), 1.);
    hypothesis2D::sigmaSpan = atof(argv[4]);

    image_loader<unsigned char>::set il_param(argv[1], 99, ".png");
    correlation_observer2D<hypothesis2D, unsigned char>::set obs_param(atoi(argv[2]), atoi(argv[3]), 40, 40);
    analyse_result_blob<hypothesis2D, unsigned char>::set analyse_param(.55 ,10 ,7);
    
    hypothesis_filter<image_loader, correlation_observer2D, analyse_result_blob,
      hypothesis2D, p_condensation, no_compensation> hf(init, atoi(argv[5]), il_param, obs_param, analyse_param);

    //tracking

    int imageNumber = 0;
    
    while(true)
      {
	try{hf.next();}
	catch (mimasexception &e){
	  std::cerr<<e.what()<<std::endl;  
	  return 0;
	}
	
	std::vector<hypothesis2D> estimatedLocation;
	hf.filter();

	//find the best peak and do a partial reinitialisation around it
	std::vector<hypothesis2D>::iterator it_h;
	///if an hypothesis has an excellent result we did a partial_reset around it

	for(it_h = hf.hypotheses.begin(); it_h != hf.hypotheses.end(); it_h++)
	  if (it_h->weight > .85)
	    {
	      hf.partial_reset(*it_h, .60, .2 , .3, true);
	      std::cerr<<"One partial reinitialisation done" <<std::endl;
	    }
	
	estimatedLocation = hf.analyse();
	
	hf.draw(estimatedLocation);//from Observer

	//----------save on a file--------//	
	std::ostringstream str;
	str << "results/pipetteBlobReinit" << std::setw(3) << std::setfill( '0' ) <<  imageNumber++
	    << ".ppm";
	std::ofstream filteredOut(str.str().c_str(), std::ios::binary); 
	filteredOut << hf.image;
	//---------------------------------//

	
	display << hf.image;//<< mimas::pause();
      }
	 
    
  } catch ( exception &e )
    {
      cerr << "An exception occurred: " <<  e.what() << endl;
      retVal = 1;
    };
  return retVal;
}
