/**
 * 2-D of freedom 
 *
 *
 * Time-stamp: <05/10/19 21:06:11 engmb>
 * Copyright (C) 2005 by Manuel Boissenin
 *
 * @author Manuel Boissenin
 * @date 17/01/2005
 ***********************************************************************/

#ifndef NDEBUG
#include <boost/multi_array.hpp>
#endif
#include <fstream>
#include "stencil_tracking.h"
#include "gnuplot.h"
#include "image_fileoutput.h"

stencil_tracking::stencil_tracking(const std::vector<Feature> &feature_images,
				   mimas::x11_display *disp,
				   int x_min, int x_max, int x_sub,
				   int y_min, int y_max, int y_sub,
				   float theta_min, float theta_max, int theta_sub,
				   int nb_thread, std::string log_name):
  x_shift_min(x_min), y_shift_min(y_min), x_sub(x_sub), y_sub(y_sub), theta_sub(theta_sub),
  feature_images(feature_images),
  transformation_space(boost::extents[x_sub][y_sub][theta_sub]),
  switch_p(0), nb_thread(nb_thread),barrier1(nb_thread + 1), barrier2(nb_thread + 1),
  current_features_to_track(*(new std::vector<Feature>))  //not elegant + memory leak...
  , display(disp), log_name(log_name),image_number_tracked(0) 
{
  current_stencil = 0;
  histo_display_time = 2000;
  std::vector<Feature>::const_iterator it = feature_images.begin(); 
  x_model_min = (*it)(0); 
  int x_model_max = (*it)(0);
  y_model_min = (*it)(1);
  int y_model_max = (*it)(1);

  for(;it != feature_images.end(); it++)
    {
      if((*it)(0) < x_model_min)
	x_model_min = (*it)(0);
      if((*it)(0) > x_model_max)
	x_model_max = (*it)(0);
      if((*it)(1) < y_model_min)
	y_model_min = (*it)(1);
      if((*it)(1) > y_model_max)
	y_model_max = (*it)(1);
    }
  
  model_width = x_model_max - x_model_min;
  model_height = y_model_max - y_model_min;
 
  /*
  std::cerr <<"Feature expressed relatively to top lef-corner features"<<std::endl; 
  std::cerr << model_width <<"," <<model_height <<std::endl;
  for(std::vector<Feature>::const_iterator it = feature_images.begin(); it != feature_images.end(); it++)
    {
      std::cerr << (*it)(0) - x_model_min<< ", "<<  (*it)(1) - y_model_min <<std::endl; 
    }
  */

  /*
#ifndef NDEBUG
  std::cerr<<"Number of features(if 1 layered model): "<<feature_images[0].size() <<std::endl;
  std::cerr<<"Number of transformations(if 1 layered model): "<<x_sub * y_sub * theta_sub <<std::endl;
#endif
  */
  float x_size = float(x_max - x_min)/ x_sub; //check this is not rounded
  float y_size = float(y_max - y_min)/ y_sub; 
  float theta_size = float(theta_max - theta_min)/ theta_sub; 

  if(x_size < 1. || y_size < 1.)
    std::cerr<<"CAREFULL, YOU HAVE SPECIFIED A BIN SIZE SMALLER THAN A PIXEL FOR "
	     <<"YOUR TRANSLATIONS SOME OF YOUR STENCIL WILL BE EMPTY SET"<<std::endl;

  std::cerr<<"Size of the subdivisions x,y : "<< x_size <<"," << y_size <<std::endl;

  //For each thread allocate space for their associate image space
  for (int i = 0; i < nb_thread; i++)
    {
      boost::multi_array<std::vector<Transformation *>,2> *image_space = 
	new boost::multi_array<std::vector<Transformation *>,2>(boost::extents[model_width + (x_max - x_min)][model_height + (y_max - y_min)]);
      image_space_vec.push_back(image_space);
    }

  image_space_I = new boost::multi_array<std::vector<Transformation *>,2>(boost::extents[model_width + (x_max - x_min)][model_height + (y_max - y_min)]);

  //boost::array<array_type::index,ndims> dims = {{4, 3, 2}};       
  //  boost::array<boost::multi_array<std::vector<Transformation *>,2>::index,2> dims = {{model_width + (x_max - x_min),model_height + (y_max - y_min)}};  
  //(*image_space_I).reshape(dims);


  
  std::cerr<<"offset values" <<x_shift_min <<", "<< x_model_min <<", "<< y_shift_min <<", "<< y_model_min <<std::endl;

  x_offset_hack_variable = x_shift_min + x_model_min;
  y_offset_hack_variable = y_shift_min + y_model_min;

  for(int i = 0; i < nb_thread; i++)
    {
      threads.create_thread(track_t(*image_space_vec[0] , transf_it, current_features_to_track, &barrier1, &barrier2, x_shift_min + x_model_min, y_shift_min + y_model_min));
    }
  

  //Fill the transformation space with its values
  for(idx_ts x = 0; x < x_sub; x++)
    for(idx_ts y = 0; y < y_sub; y++)
      for(idx_ts theta = 0; theta  < theta_sub; theta++)
	  {
	    std::vector<Interval> tmp_interv;
	    tmp_interv.push_back(Interval(x_min + x * x_size, x_min + (1+x) * x_size));
	    tmp_interv.push_back(Interval(y_min + y * y_size, y_min + (1+y) * y_size));
	    tmp_interv.push_back(Interval(theta_min + theta * theta_size, theta_min + (1+theta)*theta_size));

	    Transformation tmp_transf;
	    tmp_transf.intervals = tmp_interv;

	    transformation_space[x][y][theta] = tmp_transf; 
	  }
}
  

/*Looking at the reference table we determine again the stencils and create
  again a new reference table
  This is the 2D case first.
*/

void stencil_tracking::fill_image_space_I(Transformation *tmp_transf, float ratio)
{
  switch_p++;

  stencil_info light_stencil(tmp_transf);

  std::vector<int> histogram(500);
  for(int index = 0; index < histogram.size(); index++)
    histogram[index] = 0;

  mimas::image<unsigned char> modified_stencil;
  mimas::image<unsigned char> stencil;
  stencil.init((*image_space_vec[switch_p % nb_thread]).shape()[0] , (*image_space_vec[switch_p % nb_thread]).shape()[1]);
  modified_stencil.init((*image_space_vec[switch_p % nb_thread]).shape()[0] , (*image_space_vec[switch_p % nb_thread]).shape()[1]);

  for( std::vector<Feature>::const_iterator feature_it = feature_images.begin();
       feature_it != feature_images.end();
       feature_it++)
    {
    
      std::set<Feature, ltfeature> x_trans_rot_locus;
      //We get at least one feature (corresponding to the lower bound)
      for(int x = int(round(tmp_transf->intervals[0].lower_bound)); x < int(round(tmp_transf->intervals[0].upper_bound)); x++)
	{
	  Feature tmp_feature(2);
	  tmp_feature(0) = (*feature_it)(0) + x;
	  tmp_feature(1) = (*feature_it)(1);
	     	     
	  x_trans_rot_locus.insert(tmp_feature);
	}

      std::set<Feature, ltfeature> xy_trans_rot_locus;
      for(int y = int(round(tmp_transf->intervals[1].lower_bound)); y < int(round(tmp_transf->intervals[1].upper_bound)); y++)
	{
	  for( std::set<Feature, ltfeature>::const_iterator feat_it =
		 x_trans_rot_locus.begin();
	       feat_it != x_trans_rot_locus.end(); feat_it++)
	    {
	      Feature tmp_feature(2);
	      tmp_feature(0) = (*feat_it)(0);
	      tmp_feature(1) = (*feat_it)(1) + y;
	      xy_trans_rot_locus.insert(tmp_feature);
	    }
	}	    

      //Here we have the points of the stencil corresponding to the given transformation.
      //std::vector<int> histogram(max_number_of_elements + 1);
      
      for( std::set<Feature, ltfeature>::const_iterator feat_it = xy_trans_rot_locus.begin();
	   feat_it != xy_trans_rot_locus.end();
	   feat_it++)
	{
	  //The features are not repeated beause we are using a std::set primitive to store them.
	  //We examine here for a given stencil how many elements are associated with each position 
	  //of the stencil.

	  int x = (*feat_it)(0) - x_shift_min - x_model_min;
	  int y = (*feat_it)(1) - y_shift_min - y_model_min;
	  
	  //if(((*image_space_vec[switch_p % nb_thread])[x][y].empty()) || ((*image_space_vec[switch_p % nb_thread])[x][y].back() != tmp_transf))
	  
	  int nb_stencil_at_this_position = (*image_space_vec[switch_p % nb_thread])[x][y].size();
	  
	  //light_stencil.add_element(x, y, nb_stencil_at_this_position);
	  light_stencil.add_element(x, y, (*image_space_vec[switch_p % nb_thread])[x][y]);
	  stencil.setPixel(x, y, nb_stencil_at_this_position);
	  histogram[nb_stencil_at_this_position] += 1 ; 
	  
	}
    }

  int stencil_robustness = light_stencil.robustness(log_name);
  if (stencil_robustness < array_robustness)
    array_robustness = stencil_robustness;
  light_stencil.select_stencil_elements(ratio);
  
  //fill the improved array.
  for(std::list<stencil_element>::iterator it = light_stencil.stencil.begin(); it != light_stencil.stencil.end(); it++)
    {
      (*image_space_I)[it->x][it->y].push_back(tmp_transf);
    }

  //We then save the stencil to later be able to evaluate its robustness.
  //We could proceed the same way (instead of recalculating the shape of the stencil as we just did)
  //to obtain light stencils.

  std::pair<Transformation * , std::list<stencil_element> > my_pair(tmp_transf , light_stencil.stencil);
  light_stencils.push_back(my_pair);

      
  //show resulting stencil:

#ifndef NDEBUG
#ifndef PREPROCESSING_STAGE_NO_DEBUG 

  //PARAMETRES
  int time_waiting = 700;
  if(current_stencil%67 == 0)
    {
      for(std::list<stencil_element>::iterator it = light_stencil.stencil.begin(); it != light_stencil.stencil.end(); it++)
	{
	  modified_stencil.setPixel(it->x, it->y, it->weight);
	}
      
      mimas::image< mimas::rgba<unsigned char> > histogram_im;
      mimas::gnuplot modified_stencil_histo(histogram_im, 500,500);
      //  histo.plot_histogram(histogram, "Distribution of the number of stencils overlapping for a given stencil");
      modified_stencil_histo.plotImageProfile2(modified_stencil, "modified stencil profile");
      modified_stencil_histo.finish();
      std::stringstream file_name;
      file_name << "modified_stencil_profile_" << current_stencil << ".gp";
      std::ofstream f((log_name + file_name.str()).c_str()); f << modified_stencil_histo.getScript(); f.close();
#ifndef FAST_TESTS
      display<<mimas::image<unsigned char>(histogram_im) <<mimas::wait(time_waiting);
#endif
      //Save as an image//
      std::stringstream file_name2;
      file_name2 << log_name <<"modified_stencil_profile_" << current_stencil << ".png";
      std::ofstream ofile(file_name2.str().c_str(), std::ios::binary);
      ofile << mimas::setformat( "PNG" ) << histogram_im;
      //end Save as an image//


      mimas::gnuplot histo(histogram_im, 500,500);
      //  histo.plot_histogram(histogram, "Distribution of the number of stencils overlapping for a given stencil");
      histo.plotImageProfile2(stencil, "stencil profile");
      histo.finish();
      std::ofstream f1((log_name + std::string("stencil_profile.gp")).c_str());f1 << histo.getScript(); f1.close();
#ifndef FAST_TESTS
      display<<mimas::image<unsigned char>(histogram_im) <<mimas::wait(time_waiting);
      display<<stencil <<mimas::wait(time_waiting);
#endif
    }
#endif
#endif
  current_stencil++;
}



/*to evaluate how the stencil reduction affect the robustness.
This function has to be called after the surface reduction of the stencils
*/

void stencil_tracking::get_robustness(void)
{
  int stencil_robustness;
  //the robustness is smaller than the number of features which is likely to be smaller than 10 000 features
  int array_robustness = 123456789; 

  std::stringstream file_name;
  file_name << log_name <<"Robustness" << ".txt";
  std::ofstream ofile(file_name.str().c_str(), std::ios_base::app);
  ofile <<"Now onwards, the modified stencils robustness is considered: " << std::endl;

  for(std::vector<std::pair<Transformation *, std::list<stencil_element> > >::iterator it = light_stencils.begin(); it != light_stencils.end() ; it++)
    {
      stencil_info current_stencil(it->first);
      for(std::list<stencil_element>::iterator element_it = it->second.begin() ; element_it != it->second.end(); element_it++)
	{
	  current_stencil.add_element(element_it->x, element_it->y, (*image_space_I)[element_it->x][element_it->y]);
	}
      stencil_robustness = current_stencil.robustness(log_name);
   
      if (stencil_robustness < array_robustness)
	array_robustness = stencil_robustness;
    } 
  ofile << "TOTAL array robustness of modified stencils is: " << array_robustness<< std::endl;
}


void stencil_tracking::fill_image_space_I(float ratio){
  //We just open the file and close it to reset data from previous 
  //running of the program
  std::stringstream file_name_to_reset;
  file_name_to_reset << log_name <<"Robustness" << ".txt";
  std::ofstream ofile1(file_name_to_reset.str().c_str(), std::ios_base::trunc);
  ofile1.close();

  //we initialise array_robustness with a ridiculously high value
  array_robustness = 123456789;

  for(idx_ts x = 0; x < x_sub; x++)
    for(idx_ts y = 0; y < y_sub; y++)
      {
	Transformation *tmp_transf = &transformation_space[x][y][0];
        
	fill_image_space_I(tmp_transf, ratio);
      }

  std::stringstream file_name;
  file_name << log_name <<"Robustness" << ".txt";
  std::ofstream ofile(file_name.str().c_str(), std::ios_base::app);
  ofile <<"TOTAL initial array robustness is :" <<array_robustness <<std::endl;

  //examine the new reference array
#ifndef NDEBUG
#ifndef PREPROCESSING_STAGE_NO_DEBUG 
  //check the size of the image_space
  int nb_element = 0;
  int max_number_of_elements = 0;
  for(unsigned int x = 0; x < (*image_space_I).shape()[0]; x++)
    for (unsigned int y = 0; y < (*image_space_I).shape()[1]; y++)
      {      
	int current_number_of_elements = (*image_space_I)[x][y].size();
	if (current_number_of_elements > max_number_of_elements)
	  max_number_of_elements = current_number_of_elements;
	  
	nb_element += current_number_of_elements;
      }

  std::vector<int> histogram(max_number_of_elements + 1);
  for(int index = 0; index < histogram.size(); index++)
    histogram[index] = 0;

  for(unsigned int x = 0; x < (*image_space_I).shape()[0]; x++)
    for (unsigned int y = 0; y < (*image_space_I).shape()[1]; y++)
      {      
	int current_number_of_elements = (*image_space_I)[x][y].size();
	histogram[current_number_of_elements] += 1 ;
      }

  mimas::image< mimas::rgba<unsigned char> > histogram_im;
    
  std::cerr<<"" <<std::endl;
  mimas::gnuplot histo(histogram_im, 500,500);
  histo.plot_histogram(histogram, "Number of elements of the reference table \n having N elements in their list");
  histo.finish();
  std::ofstream f((log_name + std::string("Improved_distri.gp")).c_str()); f << histo.getScript(); f.close();
#ifndef FAST_TESTS
  display<<mimas::image<unsigned char>(histogram_im) <<mimas::wait(histo_display_time);
#endif

  //Save as an image//
  std::stringstream file_name4;
  file_name4 << log_name <<"Improved_distri" << current_stencil << ".png";
  std::ofstream ofile4(file_name4.str().c_str(), std::ios::binary);
  ofile4 << mimas::setformat( "PNG" ) << histogram_im;
  //end Save as an image//

  //Make cumulative histogram to show distribution of the array 
  std::vector<int> cummulative_histogram(max_number_of_elements + 1);

  int cumulator = 0;
  for(int index = 0; index < histogram.size(); index++)
    {
      cumulator += histogram[index];
      cummulative_histogram[index] = cumulator;
    }

  mimas::gnuplot histo1(histogram_im, 500,500);
  histo1.plot_histogram(cummulative_histogram, "Number of elements of the reference array\n having less that N position references");
  histo1.finish();
  std::ofstream f1((log_name + std::string("Improved_cummu.gp")).c_str()); f1 << histo1.getScript(); f1.close();

#ifndef FAST_TESTS
  display<<mimas::image<unsigned char>(histogram_im) <<mimas::wait(histo_display_time);      
#endif

  //Save as an image//
  std::stringstream file_name2;
  file_name2 << log_name <<"Improved_cummu" << current_stencil << ".png";
  std::ofstream ofile2(file_name2.str().c_str(), std::ios::binary);
  ofile2 << mimas::setformat( "PNG" ) << histogram_im;
  //end Save as an image//

  //Make weighted histogram to show influence on the tracking time
  std::vector<int> weighted_cummulative_histogram(max_number_of_elements + 1);
  cumulator = 0;
  for(int index = 0; index < histogram.size(); index++)
    {
      cumulator += histogram[index] * index;
      weighted_cummulative_histogram[index] = cumulator;
    }

  mimas::gnuplot histo2(histogram_im, 500,500);
  histo2.plot_histogram(weighted_cummulative_histogram, "Maximum number of vote that can be produced by the\n elements having less than N position references");
  histo2.finish();
  std::ofstream f2((log_name + std::string("Improved_votes.gp")).c_str());f2 << histo2.getScript(); f2.close(); 
#ifndef FAST_TESTS
 display<<mimas::image<unsigned char>(histogram_im) <<mimas::wait(histo_display_time);      
#endif
 //Save as an image//
 std::stringstream file_name3;
 file_name3 << log_name <<"Improved_votes" << current_stencil << ".png";
 std::ofstream ofile3(file_name3.str().c_str(), std::ios::binary);
 ofile3 << mimas::setformat( "PNG" ) << histogram_im;
 //end Save as an image//

  std::cerr<< "The image space has : "<<nb_element <<" elements"<< std::endl;
  std::cerr<< "Max number of elements is : "<<max_number_of_elements <<" elements"<< std::endl;
#endif
#endif     
}

/*
  Pseudo-code 

  fill_image_space()
  {
  for each transformation
     for each feature of the image corresponding to the transformation{
         calculate locus of point according to the sub-transformed space;      
         add transformation reference to this locus to the image_space array;
     }
  }

this function isn't called by the constructor in the hope later to serialize the
tracker (it is very long to calculate the data structure)
*/


void stencil_tracking::fill_image_space(Transformation *tmp_transf)
{
   for( std::vector<Feature>::const_iterator feature_it = feature_images.begin();
       feature_it != feature_images.end();
       feature_it++)
    {
      //calculate the locus for this feature for this transformation
      //lets start with rotation we assume !!!that
      //leap of one degre won't make us lose any pixel
      //we use a set of feature position avoiding repetition
      /*
      std::set<Feature, ltfeature> rotated_locus;

      //!!round value to find the integer box
      //	      std::vector< std::pair< float, float > >::const_iterator rotation_index = rotations.begin();
	//	      for(float t = tmp_transf->intervals[2].lower_bound; t <= tmp_transf->intervals[2].upper_bound; t++)
      
      for(std::vector< std::pair< float, float > >::const_iterator rotation_index = rotations.begin();
	  rotation_index != rotations.end(); rotation_index++)
	{
	  Feature tmp_feature(2);
	  tmp_feature(0) =
	    (int)( (*feature_it)(0) * rotation_index->first -
		   (*feature_it)(1) * rotation_index->second );
	  tmp_feature(1) =
	    (int)( (*feature_it)(0) * rotation_index->second +
		   (*feature_it)(1) * rotation_index->first );
	  rotated_locus.insert( tmp_feature );
	  //		  rotation_index++;
	}
      */
    
      std::set<Feature, ltfeature> x_trans_rot_locus;
      //We get at least one feature (corresponding to the lower bound)
      for(int x = int(round(tmp_transf->intervals[0].lower_bound)); x < int(round(tmp_transf->intervals[0].upper_bound)); x++)
	{
	  Feature tmp_feature(2);

	  tmp_feature(0) = (*feature_it)(0) + x;
	  tmp_feature(1) = (*feature_it)(1);
	     	     
	  x_trans_rot_locus.insert(tmp_feature);
	}

      std::set<Feature, ltfeature> xy_trans_rot_locus;
      for(int y = int(round(tmp_transf->intervals[1].lower_bound)); y < int(round(tmp_transf->intervals[1].upper_bound)); y++)
	{
	  for( std::set<Feature, ltfeature>::const_iterator feat_it =
		 x_trans_rot_locus.begin();
	       feat_it != x_trans_rot_locus.end(); feat_it++)
	    {
	      Feature tmp_feature(2);
	      tmp_feature(0) = (*feat_it)(0);
	      tmp_feature(1) = (*feat_it)(1) + y;
	      xy_trans_rot_locus.insert(tmp_feature);
	    }
	}

#ifndef NDEBUG	
      //display features to check that we have the right points
      /*  multi_array<bool,2> model(extents[image_space.shape()[1]][image_space.shape()[0]]);
	  for(std::set<Feature, ltfeature>::const_iterator feat_it = xy_trans_rot_locus.begin(); feat_it != xy_trans_rot_locus.end(); feat_it++)
	  {
	  model[image_space.shape()[1]/2 -(*feat_it)(1)][(*feat_it)(0) + image_space.shape()[0] /2 ] = true; 
	  }
	  display << multi_array_to_image( tobw( model ) )<<mimas::delay(500);
      */
#endif
      //for parallelisation we fill a stencil for each image space
      //fill the image space with the pointer to the transformation
	    
	      
      for( std::set<Feature, ltfeature>::const_iterator feat_it = xy_trans_rot_locus.begin();
	   feat_it != xy_trans_rot_locus.end();
	   feat_it++)
	{
	  //We are using an image to store the stencil
	  //coordinate must be shift so they are both positive
	  /*	  int x = (*feat_it)(0) + (*image_space_vec[0]).shape()[0]/2;
	  int y = (*feat_it)(1) + (*image_space_vec[0]).shape()[1]/2;
	  */
	  
	  int x = (*feat_it)(0) - x_shift_min - x_model_min;
	  int y = (*feat_it)(1) - y_shift_min - y_model_min;

	  //  std::cerr<<"feature position in array"<< x << " " << y <<std::endl; 
	  
	  //we don't add twice the same transform even if
	  //the same sub-transformation bring two points
	  //to the same place
	  if(((*image_space_vec[switch_p % nb_thread])[x][y].empty()) || ((*image_space_vec[switch_p % nb_thread])[x][y].back() != tmp_transf))
	    {
	      (*image_space_vec[switch_p % nb_thread])[x][y].push_back(tmp_transf);
	    }	      
	}
      
    }
}

void stencil_tracking::fill_image_space(void){
#ifndef NDEBUG
#ifndef PREPROCESSING_STAGE_NO_DEBUG
  unsigned n = 0;
#endif
#endif

  for(idx_ts theta = 0; theta  < theta_sub; theta++) {

    rotations.clear();
    // Pre-compute rotations to save time.
    Interval *angles = &(transformation_space[0][0][theta].intervals[2]);
    for ( float t = angles->lower_bound; t<= angles->upper_bound; t++ ) {
      float radian = t * M_PI / 180.0;
      rotations.push_back( std::make_pair( cos( radian ), sin( radian ) ) );
    };
    for(idx_ts x = 0; x < x_sub; x++)
      for(idx_ts y = 0; y < y_sub; y++)
        {

	  switch_p++;
          Transformation *tmp_transf = &transformation_space[x][y][theta];
          // for(unsigned int n=0; n < transformation_space.num_elements(); n++) {
	  // Transformation tmp_transf =  transformation_space.data()[n];
	  //intervals[3] correspond to the number of the image.

	  fill_image_space(tmp_transf);
	
	  
#ifndef NDEBUG	
#ifndef PREPROCESSING_STAGE_NO_DEBUG
	  n++;
	  if(!(n%3)) //prime number to avoid get trapped on a pattern corresponding to the same model images.
	    {		
	      //display image space for this transformation
	      //	      boost::multi_array<bool,2> transf_im(boost::extents[image_space_vec[0]->shape()[1]][image_space_vec[0]->shape()[0]]);
	      mimas::image<unsigned char> transf_im;
	      transf_im.init(image_space_vec[0]->shape()[0],image_space_vec[0]->shape()[1]);

	      for(idx_is x = 0; (unsigned int)(x) < image_space_vec[0]->shape()[0] ; x++)
		for(idx_is y = 0;(unsigned int) (y) < image_space_vec[0]->shape()[1] ; y++)
		  if((!(*image_space_vec[n % nb_thread])[x][y].empty()) && ((*image_space_vec[n % nb_thread])[x][y].back() == tmp_transf ))
		    {
		      transf_im.pixel(x,y) = 255; 
		    }
#ifndef FAST_TESTS
	      display <<  transf_im << mimas::delay(100);
#endif
	    }	  
#endif
#endif	 
	}
  }

#ifndef NDEBUG 
#ifndef PREPROCESSING_STAGE_NO_DEBUG
  //check the size of the image_space
  int nb_element = 0;
  for (int i = 0; i < nb_thread; i++)
    {
      int max_number_of_elements = 0;
      for(unsigned int x = 0; x < image_space_vec[i]->shape()[0]; x++)
	for (unsigned int y = 0; y < image_space_vec[i]->shape()[1]; y++)
	  {      
	    int current_number_of_elements = (*image_space_vec[i])[x][y].size();
	    if (current_number_of_elements > max_number_of_elements)
	      max_number_of_elements = current_number_of_elements;

	    nb_element += current_number_of_elements;
	  }

      std::vector<int> histogram(max_number_of_elements + 1);
      for(int index = 0; index < histogram.size(); index++)
	histogram[index] = 0;

      for(unsigned int x = 0; x < image_space_vec[i]->shape()[0]; x++)
	for (unsigned int y = 0; y < image_space_vec[i]->shape()[1]; y++)
	  {      
	    int current_number_of_elements = (*image_space_vec[i])[x][y].size();
	    histogram[current_number_of_elements] += 1 ;
	  }

      mimas::image< mimas::rgba<unsigned char> > histogram_im;

      std::cerr<<"" <<std::endl;
      mimas::gnuplot histo(histogram_im, 500,500);
      histo.plot_histogram(histogram, "Number of elements of the reference table \n having N elements in their list");
      histo.finish();
      std::ofstream f((log_name + std::string("distri.gp")).c_str()); f << histo.getScript(); f.close();     
 
#ifndef FAST_TESTS
     display<<mimas::image<unsigned char>(histogram_im) <<mimas::wait(histo_display_time);
#endif

      //Make cumulative histogram to show distribution of the array 
      std::vector<int> cummulative_histogram(max_number_of_elements + 1);

      int cumulator = 0;
      for(int index = 0; index < histogram.size(); index++)
	{
	  cumulator += histogram[index];
	  cummulative_histogram[index] = cumulator;
	}

      mimas::gnuplot histo1(histogram_im, 500,500);
      histo1.plot_histogram(cummulative_histogram, "Number of elements of the reference array\n having less that N position references");
      histo1.finish();
      std::ofstream f2((log_name + std::string("cummu.gp")).c_str()); f2 << histo1.getScript(); f2.close();      
#ifndef FAST_TESTS      
      display<<mimas::image<unsigned char>(histogram_im) <<mimas::wait(histo_display_time);      
#endif
      //Make weighted histogram to show influence on the tracking time
      std::vector<int> weighted_cummulative_histogram(max_number_of_elements + 1);
      cumulator = 0;
      for(int index = 0; index < histogram.size(); index++)
	{
	  cumulator += histogram[index] * index;
	  weighted_cummulative_histogram[index] = cumulator;
	}

      mimas::gnuplot histo2(histogram_im, 500,500);
      histo2.plot_histogram(weighted_cummulative_histogram, "Maximum number of vote that can be produced by the\n elements having less than N position references");
      histo2.finish();
      std::ofstream f3((log_name + std::string("votes.gp")).c_str()); f3 << histo2.getScript(); f3.close();     
#ifndef FAST_TESTS
      display<<mimas::image<unsigned char>(histogram_im) <<mimas::wait(histo_display_time);      
#endif
      std::cerr<< "The image space for the thread "<< i <<" has : "<<nb_element <<" elements"<< std::endl;
      std::cerr<< "Max number of elements for the thread "<< i <<" is : "<<max_number_of_elements <<" elements"<< std::endl;
      nb_element = 0;
    }
#endif
#endif     
}

Transformation stencil_tracking::track(std::vector<Feature> &features)
{
  image_number_tracked ++;
#ifndef NDEBUG
  std::cerr<<"Begining of the tracking function"<<std::endl; 
  mimas::timer t;    
#endif

  //reinitialize the number of vote
  for(unsigned int n = 0; n < transformation_space.num_elements(); n++)
    transformation_space.data()[n].votes = 0;

  current_features_to_track = features;
  //make the threads wait for the previous line to be done
  barrier1.wait();
  //Synchonise the treads so they all finish their job before
  //looking fo a maximum
  barrier2.wait();

  std::vector<int> nb_votes;
  for(unsigned int n = 0; n < transformation_space.num_elements(); n++)
    nb_votes.push_back(transformation_space.data()[n].votes);

  Transformation max_element = *(std::max_element(transformation_space.data(),
						  transformation_space.data()+ transformation_space.num_elements(),
						  lttransf()));

#ifndef NDEBUG
#ifndef TRACKING_NO_DEBUG 
  std::cerr<<"from track()"<<std::endl;
#ifndef FAST_TESTS
  max_element.debug();
#endif
  std::cerr<<"Number of millisecond to track one object in this image: "<< t.elapsed("ms")<< std::endl;

  mimas::image< mimas::rgba<unsigned char> > im;
  mimas::gnuplot histo2(im, 500,500);
  histo2.plot_histogram(nb_votes, "Number of votes received by a stencil");
  std::ostringstream str;
  str << log_name << std::setw(3) << std::setfill( '0' ) <<  image_number_tracked
      <<std::string("voting.gp") ;
  histo2.finish();
  std::ofstream f1(str.str().c_str()); f1 << histo2.getScript(); f1.close();
#ifndef FAST_TESTS
  display<< mimas::image< unsigned char >(im);
#endif

#endif
#endif
  return max_element;
}


Transformation stencil_tracking::track_I(std::vector<Feature> &features)
{
  image_number_tracked ++;
  
  //reinitialize the number of vote
  for(unsigned int n = 0; n < transformation_space.num_elements(); n++)
    transformation_space.data()[n].votes = 0;
 
  current_features_to_track = features;

#ifndef NDEBUG 
  double time;
  mimas::timer t;    
#endif
 
  for( std::vector<Feature>::const_iterator feature_it = features.begin(); feature_it != features.end(); feature_it++)
    {
      Feature temp(2);
 	      
      //Consider only features that are within the object movement bounds 
      temp(0) = (*feature_it)(0) - x_offset_hack_variable;
      temp(1) = (*feature_it)(1) - y_offset_hack_variable;
       
      if(temp(0)>=0 && temp(0) < image_space_I->shape()[0] && temp(1)>=0 && temp(1) < image_space_I->shape()[1])
 	for(transf_it = (*image_space_I)[temp(0)][temp(1)].begin();
 	    transf_it != (*image_space_I)[temp(0)][temp(1)].end();
 	    transf_it++)
 	  (*transf_it)->votes++;
    }
 
#ifndef NDEBUG
  time =  t.elapsed("ms");
#endif
  //We only measure here the voting time, total tracking time must include the time to look for the maximum
 
  std::vector<int> nb_votes;
  for(unsigned int n = 0; n < transformation_space.num_elements(); n++)
    nb_votes.push_back(transformation_space.data()[n].votes);
  
  Transformation *max_element_p = std::max_element(transformation_space.data(),
  						  transformation_space.data()+ transformation_space.num_elements(),
  						  lttransf());


#ifndef NDEBUG
  double time2 =  t.elapsed("ms");
  //#ifndef FAST_TESTS
  std::cerr<<"from track_I():"<<std::endl;
  std::cerr<<"Number of millisecond to track one object in this image: "<< time << std::endl;
  std::cerr<<"Number of millisecond to track one object in this image: "<< time2 << std::endl;
  //#endif

#ifndef FAST_TESTS
  max_element_p->debug();
#endif

  std::stringstream file_name;
  file_name << log_name <<"tracking_times" << ".txt";
  std::ofstream ofile(file_name.str().c_str(), std::ios_base::app);
  ofile << time <<std::endl;
  ofile.close();

  //Here we are evaluating the proportion of wining stencil elements that got a hit.
  //We first get the the transformation element
  int number_of_stencil_elements = 0;
  for(std::vector<std::pair<Transformation *, std::list<stencil_element> > >::iterator it =  light_stencils.begin(); it != light_stencils.end(); it++)
    {
      if(it->first == max_element_p)
	  {
	    number_of_stencil_elements = it->second.size();
	  }
    }

  int number_of_votes = max_element_p->votes;
#ifndef FAST_TESTS
  std::cout << "Proportion of stencil element hitten for the winning stencil: " << float(number_of_votes)/ float(number_of_stencil_elements)<<std::endl;
#endif
  //We now consider the second stencil that got the maximum number of votes:

  int max = 0;
  Transformation *second_max_element_p;
  for(Transformation * it = transformation_space.data(); 
      it != transformation_space.data() + transformation_space.num_elements(); it++) 
    {
      if(it->votes >max && it != max_element_p)
	{
	  max = it->votes;
	  second_max_element_p = it;
	}
    }
  
  int number_of_stencil_elements2 = 0;
  for(std::vector<std::pair<Transformation *, std::list<stencil_element> > >::iterator it =  light_stencils.begin(); it != light_stencils.end(); it++)
    {
      if(it->first == second_max_element_p)
	  {
	    number_of_stencil_elements2 = it->second.size();
	  }
    }

  int number_of_votes2 = second_max_element_p->votes;

#ifndef FAST_TESTS
  std::cout << "Proportion of stencil element hitten for the second stencil getting more votes: " << float(number_of_votes2)/ float(number_of_stencil_elements2)<<std::endl;
#endif

  std::stringstream file_name2;
  file_name2 << log_name <<"Second_stencil_hit_percentage.dat";
  std::ofstream ofile2(file_name2.str().c_str(), std::ios_base::app);
  ofile2 << float(number_of_votes2)/ float(number_of_stencil_elements2) <<std::endl;
  ofile2.close();


#ifndef TRACKING_NO_DEBUG
  mimas::image< mimas::rgba<unsigned char> > im;
  mimas::gnuplot histo2(im, 500,500);
  histo2.plot_histogram(nb_votes, "Number of votes received by a stencil");
  histo2.finish();

  if((image_number_tracked %20) == 0)  //to save a selection of the graphs 
    {
      std::ostringstream str;
      str << log_name << std::setw(3) << std::setfill( '0' ) <<  image_number_tracked
	  <<std::string("votingI.gp");
      std::ofstream f1(str.str().c_str()); f1 << histo2.getScript(); f1.close();
      //Save as an image//
      std::stringstream file_name;
      file_name << log_name <<"votingI" << image_number_tracked << ".png";
      std::ofstream ofile(file_name.str().c_str(), std::ios::binary);
      ofile << mimas::setformat( "PNG" ) << im;
      //end Save as an image//
    }
#ifndef FAST_TESTS
  display<< mimas::image< unsigned char >(im);
#endif

#endif
#endif

  return *max_element_p;
}
