//
// gradient filters
// stumeikle Sat May  4 20:05:34 2002
//
//


#ifndef FILTER_GRAD_H
#define FILTER_GRAD_H

#include <boost/shared_array.hpp>
#include <string>
#include <vector>
#include "image.h"
#include "filter.h"

namespace mimas {
/*  General image filter routines... continued. Gradient filters are placed in
    this file. Chose to break up the original filter.h file to keep the
    file sizes down.
    $Header: /cvs/mimas/include/filter_grad.h,v 1.2 2002/11/19 11:40:20 bpa Exp
 */
/** @addtogroup gradientFilters
    @{ */
template<typename T>
void filter<T>::gradX( const image<T>& input, image<T>& output )
{
    if ( &output != &input )
      output.init( input.getWidth(), input.getHeight());

    boost::shared_array< T > row( new T[input.getWidth()] );

    for(int y=0;y<input.getHeight();++y)
    {
	for( int x=1;x<(input.getWidth()-1);++x)
	{
	    row[x]=(T)(((double)input.pixel(x+1,y) - (double)input.pixel(x-1,y))/2.0);
	}

   int x = input.getWidth() - 1;
	row[0] = input.pixel(1,y) - input.pixel(0,y);
	row[x] = input.pixel(x,y) - input.pixel(x-1,y);
        std::copy( row.get(), row.get() + input.getWidth(), output.rawData() + y * output.getWidth() );
    }

    // delete[] row;
}

template<typename T>
void filter<T>::gradY( const image<T>& input, image<T>& output )
{
    if ( &output != &input )
	output.init( input.getWidth(), input.getHeight());

    boost::shared_array< T > col( new T[input.getHeight()] );

    for(int x=0;x<input.getWidth();++x)
    {
	for(int y=1;y<(input.getHeight()-1);++y)
	{
	    col[y]=(input.pixel(x,y+1) - input.pixel(x,y-1))/2;
	}

   int y = input.getHeight() - 1;
	col[0] = input.pixel(x,1) - input.pixel(x,0);
	col[y] = input.pixel(x,y) - input.pixel(x,y-1);

	for(int y=0;y<input.getHeight();++y)
	    output.pixel(x,y)=col[y];
    }
}

///@}
}

#endif
