#ifndef PRIMITIVE_H
#define PRIMITIVE_H

#include "mimasexception.h"

namespace mimas {
/**
 * A pixel.
 */
struct pixel {
 
  pixel operator=(const pixel &rhs)
  {
    this->x = rhs.x;
    this->y = rhs.y;
    return *this;
  }

  pixel operator+(const pixel &rhs) const
  {
    pixel result;
    result.x = this->x + rhs.x;
    result.y = this->y + rhs.y;
    return result;
  }

  unsigned int x,y;
  pixel(int x = 0, int y = 0) throw(mimasexception)
  {
    if (x < 0 || y < 0)
       {
	 mimasexception e;
	 e << "Pixel must have positive coordinates" ;              
	 throw e;  
       }
    else
      {
	this->x = x;
	this->y = y;
      }
  }
};


/**
 * A double-precision pixel. Useful for sub-pixel accuracy.
 */
struct pixel_f {
  double x, y;
  pixel_f( double x = 0, double y = 0): x(x), y(y){ assert(x >=0 && y>= 0); }
};

/**
 * A int point.
 */
struct point {
  point operator=(const point &rhs)
  {
    this->x = rhs.x;
    this->y = rhs.y;
    return *this;
  }

  point operator+(const point &rhs) const
  {
    point result;
    result.x = this->x + rhs.x;
    result.y = this->y + rhs.y;
    return result;
  }
  int x,y;
  point(int x = 0, int y = 0): x(x), y(y){}
};

/**
 * A float point.
 */
struct point_f {
  double x,y;
  point_f(double x = 0, double y = 0): x(x), y(y){}
};

/**
 * A segment (made of 2 points).
 */
template< typename T >
struct segment {
  T p1,p2;
  segment(T &p1, T &p2): p1(p1), p2(p2){}
};

}    
#endif
