#ifndef X11_WINDOW_H
#define X11_WINDOW_H

#include "mimasconfig.h"

#ifdef HAVE_OPENGL

#include <boost/shared_ptr.hpp>
#define GLX_GLXEXT_PROTOTYPES
#include <GL/glx.h>
#include <map>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "object.h"
#include "mimasexception.h"

namespace mimas {

class x11_display;

/** Class for opening X11 windows.
    @author Jan Wedekind (jan at wedesoft.de)
    @date Fri Apr 29 23:10:01 UTC 2005 */
class x11_window: public object
{
  friend class x11_display;
 public:
  ///
  x11_window( x11_display *_display, int _width, int _height,
	      bool depthBuffer = false );
  ///
  virtual ~x11_window(void);
  ///
  int get_width(void) const { return width; }
  ///
  int get_height(void) const { return height; }
  ///
  void set_title( const char *_title );
  ///
  XVisualInfo *get_visual_info(void) const { return visualInfo; }
  ///
  x11_display *get_display(void) { return display; }
  ///
  Window get_window(void) { return window; }
  /// Resize the window.
  void resize( int _width, int _height ) throw (mimasexception);
  ///
  void repaint(void);
 protected:
  /** Handle X11 events.
      The default-implementation just filters for exposure events and
      forwards them to \c paintEvent. */
  virtual void handle_event( XEvent &event ) throw (mimasexception);
  ///
  virtual void paint_event(void) throw (mimasexception) = 0;
  ///
  virtual void key_event( XKeyEvent &xkey ) throw (mimasexception);
  ///
  static Bool waitForNotify( Display *, XEvent *e, char *arg );
  ///
  x11_display *display;
  ///
  XVisualInfo *visualInfo;
  /// Colormap of X11-visual.
  Colormap colorMap;
  /// X11 window.
  Window window;
  ///
  int width;
  ///
  int height;
};

///
typedef boost::shared_ptr< x11_window > x11_window_ptr;

};

#endif

#endif

