#ifndef __XML_ERRORHANDLER_H
#define __XML_ERRORHANDLER_H

#include "mimasconfig.h"

#ifdef HAVE_XALAN

#include <sax/ErrorHandler.hpp>
#include <string>

#ifdef XERCES_HAS_CPP_NAMESPACE
using namespace XERCES_CPP_NAMESPACE;
#endif
#ifdef XALAN_HAS_CPP_NAMESPACE
using namespace XALAN_CPP_NAMESPACE;
#endif

namespace mimas {

/** @addtogroup xml
    @{ */
/** Error handler for interfacing with Xalan-library.
    This class is for internal use only. 
    @see xml_document
    @author Jan Wedekind (jan at wedesoft.de)
    @date Wed Jan 12 13:53:41 UTC 2004 */
class xml_errorhandler: public ErrorHandler
{
public:
  /// Constructor.
  xml_errorhandler(void);
  /** Call-back function for warnings.
      Does nothing. In debug-mode information is printed to stderr.
      @param toCatch Exception object. */
  void warning(const SAXParseException &toCatch);
  /** Call-back function for errors.
      \c triggerError is called.
      @param toCatch Exception object. */
  void error(const SAXParseException &toCatch);
  /** Call-back function for fatal errors.
      \c triggerError is called.
      @param toCatch Exception object. */
  void fatalError(const SAXParseException &toCatch);
  /** Clear error-information to reuse handler.
      Clears the message-buffer and resets \c errorOccurred. */
  void resetErrors(void);
  /** Get error-status.
      @return \c true indicates an error has occurred. */
  bool getErrorOccurred(void) const { return errorOccurred; }
  /** Get error-messages.
      @return String containing the error-message(s). */
  const std::string &getErrorMessage(void) const { return message; }
protected:
  /** Change state and store error-description.
      \c errorOccurred is set to \c true and the description of the error
      is added to the message-buffer. */
  void triggerError( const std::string &type,
                     const SAXParseException &toCatch );
  /** Extract textual description of error.
      The information provided by the exception object is converted to a
      textual error-description.
      @param type Type of error (f.e. "Error" or "Fatal error").
      @param toCatch Exception object. */
  std::string createMessage( const std::string &type,
                             const SAXParseException &toCatch );
  /// Boolean indicating, wether an error has occurred.
  bool errorOccurred;
  /// Message-buffer with textual description of errors.
  std::string message;
};

///@}

};

#endif

#endif
